/*
 * Copyright (C) 2019 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 */
package org.securedroneid.android.bluetooth;

import android.annotation.TargetApi;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.le.BluetoothLeScanner;
import android.bluetooth.le.ScanCallback;
import android.bluetooth.le.ScanFilter;
import android.bluetooth.le.ScanRecord;
import android.bluetooth.le.ScanResult;
import android.bluetooth.le.ScanSettings;
import android.content.Context;
import android.os.Build;
import android.os.ParcelUuid;
import android.util.Log;

import org.securedroneid.android.log.LogEntry;
import org.securedroneid.android.log.LogMessageEntry;
import org.securedroneid.android.log.LogWriter;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.UUID;

public class BluetoothScanner {
    private static final String TAG = "BluetoothManager";

    private final OpenDroneIdDataManager dataManager;
    private LogWriter logger;
    private BluetoothAdapter bluetoothAdapter;
    private BluetoothLeScanner bluetoothLeScanner;

    private Context context;

    public BluetoothScanner(Context context, OpenDroneIdDataManager dataManager) {
        this.dataManager = dataManager;
        this.context = context;

        Object object = context.getSystemService(Context.BLUETOOTH_SERVICE);
        if (object == null)
            return;
        bluetoothAdapter = ((android.bluetooth.BluetoothManager) object).getAdapter();
    }

    public void setLogger(LogWriter logger) { this.logger = logger; }

    private static String dumpBytes(byte[] bytes) {
        return LogEntry.toHexString(bytes, bytes.length);
    }

    public BluetoothAdapter getBluetoothAdapter() {
        return bluetoothAdapter;
    }

    private final ScanCallback scanCallback = new ScanCallback() {

        @Override
        public void onScanResult(int callbackType, ScanResult result) {
            ScanRecord scanRecord = result.getScanRecord();
            if (scanRecord == null)
                return;
            byte[] bytes = scanRecord.getBytes();

            String addr = result.getDevice().getAddress().substring(0, 8);
            int advertiseFlags = scanRecord.getAdvertiseFlags();
            int rssi = result.getRssi();
            String string = String.format(Locale.US, "scan: addr=%s flags=0x%02X rssi=% d, len=%d",
                    addr, advertiseFlags, rssi, bytes != null ? bytes.length : -1);

            String transportType = "BT4";
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O && bluetoothAdapter.isLeCodedPhySupported()) {
                if (result.getPrimaryPhy() == BluetoothDevice.PHY_LE_CODED)
                    transportType = "BT5";
            }

            LogMessageEntry logMessageEntry = new LogMessageEntry();
            try {
                dataManager.receiveDataBluetooth(bytes, result, logMessageEntry, transportType);
            } catch(Exception e) {
                 //Uncomment this to see popup exceptions
                /*StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                String exceptionAsString = sw.toString();
                AlertDialog.Builder alertDialogBuilder = new AlertDialog.Builder(
                        context);

                // set title
                alertDialogBuilder.setTitle("Exception");

                // set dialog message
                alertDialogBuilder
                        .setMessage(exceptionAsString)
                        .setCancelable(false)
                        .setNegativeButton("No",new DialogInterface.OnClickListener() {
                            public void onClick(DialogInterface dialog,int id) {
                                // if this button is clicked, just close
                                // the dialog box and do nothing
                                dialog.cancel();
                            }
                        });

                // create alert dialog
                AlertDialog alertDialog = alertDialogBuilder.create();

                // show it
                alertDialog.show();*/
            }

            StringBuilder csvLog = logMessageEntry.getMessageLogEntry();
            if (logger != null)
                logger.logBluetooth(logMessageEntry.getMsgVersion(), result, transportType, csvLog);

            Log.w(TAG, "onScanResult: " + string);
            if (bytes != null) {
                Log.w(TAG, "-- bytes: " + dumpBytes(bytes));
            }
        }

        @Override
        public void onBatchScanResults(List<ScanResult> results) {
        }

        @Override
        public void onScanFailed(int errorCode) {
        }
    };

    /* OpenDroneID Bluetooth beacons identify themselves by setting the GAP AD Type to
     * "Service Data - 16-bit UUID" and the value to 0xFFFA for ASTM International, ASTM Remote ID.
     * https://www.bluetooth.com/specifications/assigned-numbers/ -> "Generic Access Profile"
     * https://www.bluetooth.com/specifications/assigned-numbers/ -> "16-bit UUIDs"
     * Vol 3, Part B, Section 2.5.1 of the Bluetooth 5.1 Core Specification
     * The AD Application Code is set to 0x0D = Open Drone ID.
     */
    private static final UUID SERVICE_UUID = UUID.fromString("0000fffa-0000-1000-8000-00805f9b34fb");
    private static final ParcelUuid SERVICE_pUUID = new ParcelUuid(SERVICE_UUID);
    private static final byte[] OPEN_DRONE_ID_AD_CODE = new byte[]{(byte) 0x0D};

    @TargetApi(Build.VERSION_CODES.O)
    public void startScan() {
        if (bluetoothAdapter == null)
            return;

        Log.d(TAG, ">>>> startScan");

        // Test messages to test app in AndroidStudio
        /*byte[][] messages = {
                {0x1e,0x16, (byte) 0xfa, (byte) 0xff,0x0d,0x00,0x00,0x4f,0x20,0x01,0x00,0x3f, (byte) 0xfe,0x03,0x0a,0x05,0x03,0x39, (byte) 0xef, (byte) 0xb2, (byte) 0x9f,0x4e, (byte) 0x9d,0x61,0x00,0x00,0x00,0x00,0x00,0x00,0x00}
                /*{0x1e, 0x16, (byte) 0xfa, (byte) 0xff, 0x0d, 0x00, 0x20, 0x50, 0x06, (byte) 0x89, 0x24, (byte) 0xed, 0x3d, 0x09, 0x01, 0x70, 0x20, 0x39, 0x09, 0x70, (byte) 0xa5, 0x1b, 0x0b, 0x20, 0x01, 0x00, 0x3f, (byte) 0xfe, 0x00, 0x14, 0x05},
                {0x1e, 0x16, (byte) 0xfa, (byte) 0xff, 0x0d, 0x00, 0x20, 0x51, (byte) 0xd1, (byte) 0xee, 0x02, (byte) 0x93, (byte) 0xb2, 0x1d, 0x63, (byte) 0xf3, (byte) 0xed, 0x02, (byte) 0x8d, 0x1f, (byte) 0xc2, 0x48, (byte) 0xfa, (byte) 0x93, 0x0a, 0x1b, (byte) 0x9e, (byte) 0xda, (byte) 0xf3, (byte) 0xf4, (byte) 0xc4},
                {0x1e, 0x16, (byte) 0xfa, (byte) 0xff, 0x0d, 0x00, 0x20, 0x52, 0x56, 0x2b, (byte) 0xb3, 0x64, (byte) 0x87, (byte) 0xe5, (byte) 0xee, (byte) 0x9b, (byte) 0xf2, 0x14, (byte) 0xbe, (byte) 0xab, 0x44, (byte) 0xd9, (byte) 0x88, (byte) 0xad, (byte) 0xb0, 0x20, 0x01, 0x00, 0x3f, (byte) 0xfe, 0x00},
                {0x1e, 0x16, (byte) 0xfa, (byte) 0xff, 0x0d, 0x00, 0x20, 0x53, 0x14, 0x05, (byte) 0xe2, 0x38, (byte) 0xe2, (byte) 0xbc, (byte) 0xf4, 0x01, (byte) 0xdc, 0x2d, (byte) 0xd8, (byte) 0xaa, 0x42, 0x39, (byte) 0xd5, 0x52, (byte) 0x9e, (byte) 0xd2, (byte) 0xde, 0x61, (byte) 0xc9, (byte) 0xd9, (byte) 0xc4},
                {0x1e, 0x16, (byte) 0xfa, (byte) 0xff, 0x0d, 0x00, 0x20, 0x54, 0x52, 0x6e, (byte) 0x94, (byte) 0xf4, 0x34, (byte) 0xde, 0x72, (byte) 0xe5, 0x08, (byte) 0x8d, 0x68, 0x52, 0x5d, (byte) 0xb7, 0x62, (byte) 0xf7, (byte) 0x9c, (byte) 0xb9, 0x20, 0x4e, 0x75, 0x38, (byte) 0xdf},
                {0x1e, 0x16, (byte) 0xfa, (byte) 0xff, 0x0d, 0x00, 0x20, 0x55, (byte) 0xd3, (byte) 0xfc, 0x1d, 0x13, 0x4f, 0x08, 0x60, (byte) 0x85, (byte) 0xbd, (byte) 0xea, 0x69, 0x75, (byte) 0xc1, 0x1b, 0x72, 0x4c, 0x45, (byte) 0xa1, 0x08, (byte) 0x81, 0x3c, (byte) 0xca, (byte) 0xd1},
                {0x1e, 0x16, (byte) 0xfa, (byte) 0xff, 0x0d, 0x00, 0x20, 0x56, 0x6d, 0x56, (byte) 0xe0, (byte) 0x8c, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
                {0x1e, 0x16, (byte)0xfa, (byte)0xff, 0x0d, 0x00, 0x20, 0x50, 0x05, 0x72, 0x4d, 0x2b, 0x3e, 0x09, 0x02, 0x4d, 0x2b, 0x3e, 0x09, (byte)0xcd, 0x7a, 0x50, 0x09, 0x00, 0x4f, 0x20, 0x01, 0x00, 0x3f, (byte)0xfe, 0x00},
                {0x1e, 0x16, (byte)0xfa, (byte)0xff, 0x0d, 0x00, 0x20, 0x51, 0x14, 0x05, (byte)0xd1, (byte)0xee, 0x02, (byte) 0x93, (byte)0xb2, 0x1d, 0x63, (byte)0xf3, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x01, 0x00, 0x3f, (byte)0xfe, 0x00},
                {0x1e, 0x16, (byte)0xfa, (byte)0xff, 0x0d, 0x00, 0x20, 0x52, 0x14, 0x05, (byte)0xd1, (byte)0xee, 0x02, (byte) 0x93, (byte)0xb2, 0x1d, 0x63, (byte)0xf3, (byte)0xda, (byte)0xd1, (byte)0xf9, 0x3f, 0x35, 0x67, 0x5c, 0x43, 0x7c, (byte)0xf8, 0x5b, 0x51, (byte)0x93},
                {0x1e, 0x16, (byte)0xfa, (byte)0xff, 0x0d, 0x00, 0x20, 0x53, (byte)0x99, (byte)0xbd, 0x58, (byte)0xe4, (byte)0x9b, 0x04, (byte)0x9e, (byte)0xd4, 0x75, (byte)0xc8, (byte)0xfc, 0x48, (byte)0xd6, (byte)0xfc, (byte)0xbc, (byte)0xc4, 0x7b, (byte) 0x97, 0x6f, 0x2f, 0x32, (byte)0x87, 0x04},
                {0x1e, 0x16, (byte)0xfa, (byte)0xff, 0x0d, 0x00, 0x20, 0x54, 0x02, 0x64, 0x11, (byte)0x84, 0x53, 0x2a, (byte)0x9c, (byte)0xa0, 0x65, (byte)0x9e, (byte)0xeb, 0x4d, (byte)0x89, (byte)0xb9, (byte)0xcc, (byte)0xe2, 0x4a, 0x68, 0x35, (byte)0xd9, (byte)0xb6, (byte)0xdd, (byte)0xd0},
                {0x1e, 0x16, (byte)0xfa, (byte)0xff, 0x0d, 0x00, 0x20, 0x55, (byte)0xb6, (byte)0xa7, (byte)0xc8, 0x22, 0x0a, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}

        };
        for(int i = 0; i < messages.length; i++) {
            dataManager.receiveTestData(messages[i]);
        }//*/

        bluetoothLeScanner = bluetoothAdapter.getBluetoothLeScanner();

        ScanFilter.Builder builder = new ScanFilter.Builder();
        builder.setServiceData(SERVICE_pUUID, OPEN_DRONE_ID_AD_CODE);
        List<ScanFilter> scanFilters = new ArrayList<>();
        scanFilters.add(builder.build());

        ScanSettings scanSettings = new ScanSettings.Builder()
                                    .setScanMode(ScanSettings.SCAN_MODE_LOW_LATENCY)
                                    .build();
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O &&
            bluetoothAdapter.isLeCodedPhySupported() &&
            bluetoothAdapter.isLeExtendedAdvertisingSupported()) {
            // Enable scanning also for devices advertising on an LE Coded PHY S2 or S8
            scanSettings = new ScanSettings.Builder()
                           .setScanMode(ScanSettings.SCAN_MODE_LOW_LATENCY)
                           .setLegacy(false)
                           .setPhy(ScanSettings.PHY_LE_ALL_SUPPORTED)
                           .build();
        }

        if (bluetoothLeScanner != null && bluetoothAdapter.isEnabled())
            bluetoothLeScanner.startScan(scanFilters, scanSettings, scanCallback);
    }

    public void stopScan() {
        if (bluetoothLeScanner != null && bluetoothAdapter.isEnabled()) {
            bluetoothLeScanner.stopScan(scanCallback);
        }
    }
}
