package org.securedroneid.android.bluetooth;

import static org.securedroneid.android.Constants.DET_SIZE;
import static org.securedroneid.android.Constants.HASH_SIZE;
import static org.securedroneid.android.Constants.SIGNATURE_SIZE;

import android.util.Log;

import org.bouncycastle.math.ec.rfc8032.Ed25519;
import org.securedroneid.android.Constants;
import org.securedroneid.android.data.AircraftObject;
import org.securedroneid.android.data.AuthenticationData;
import org.securedroneid.android.data.Connection;
import org.securedroneid.android.data.Identification;
import org.securedroneid.android.data.Link;
import org.securedroneid.android.data.LocationData;
import org.securedroneid.android.data.Manifest;
import org.securedroneid.android.data.OperatorIdData;
import org.securedroneid.android.data.SelfIdData;
import org.securedroneid.android.data.SystemData;
import org.securedroneid.android.data.Wrapper;

import java.io.StringWriter;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;

public class DRIPDataManager{

    static byte[] HDA_DET = new byte[]{
            0x20, 0x01, 0x00, 0x3F,
            (byte)0xFE, 0x03, 0x0A, 0x05,
            (byte)0xF9, 0x29, (byte)0x84, 0x1F,
            0x00, 0x6C, 0x18, 0x40
    }; // Assume that the HDA DET is already cached for now.

    static byte[] HDA_PUB_KEY = new byte[]{
            (byte)0x8A, (byte)0xF0, 0x5B, (byte)0xFF,
            0x6F, (byte)0xFD, (byte)0x9E, (byte)0x8F,
            0x41, 0x77, 0x4F, 0x4D,
            (byte)0xAC, (byte)0xE4, (byte)0xB1, (byte)0x98,
            (byte)0xE3, (byte)0xF8, (byte)0xE0, (byte)0x9D,
            (byte)0x92, 0x4C, 0x3F, 0x2C,
            (byte)0x94, (byte)0x88, (byte)0x29, (byte)0x8D,
            (byte)0x9D, 0x79, 0x40, (byte) 0x91
    }; // Assume that the HDA public key is already cached for now.

    static String TAG = "DRIPDataManager";

    public enum SAMType {
        None(0),
        Link(1),
        Wrapper(2),
        Manifest(3),
        Frame(4);

        SAMType(int id) { this.id = id; }
        public final int id;

        public static SAMType fromId(int id) {
            if (id == None.id) {
                return None;
            } else if (id == Link.id) {
                return Link;
            } else if (id == Wrapper.id) {
                return Wrapper;
            } else if (id == Manifest.id) {
                return Manifest;
            } else if (id == Frame.id) {
                return Frame;
            } else {
                return null;
            }
        }
    }

    public static void handleSAMMessage(ByteBuffer SAMMessageData, AircraftObject ac, AuthenticationData auth, OpenDroneIdDataManager dataManager) {

        Log.d(TAG,"Started handling SAM");

        if(SAMMessageData.remaining() == 0) {
            Log.e(TAG, "SAM Message is empty");
            return;
        }

        byte type_int = SAMMessageData.get();
        SAMType type = SAMType.fromId(type_int);

        if(type == null) {
            Log.e(TAG, "SAM Type is null (" + type_int + ")");
            return;
        }

        switch(type) {
            case Link:
            handleLink(SAMMessageData, auth, ac);
            break;
            case Wrapper:
            case Manifest:
            handleEndorsementStructure(SAMMessageData, type, ac, auth, dataManager);
            break;
            default:
        }
    }

    private static void handleLink(ByteBuffer linkData, AuthenticationData auth, AircraftObject ac) {

        int messageSize = linkData.remaining() - SIGNATURE_SIZE;
        Log.d(TAG,"Started handling link (" + messageSize + ")");

        if(messageSize <= 0) {
            Log.e(TAG, "Message size is <= 0");
            return;
        }

        byte[] entire_message = new byte[messageSize];
        linkData.slice().get(entire_message);

        int VNB = linkData.getInt();
        int VNA = linkData.getInt();

        if(!verifyTimestamps(VNB, VNA)) {
            Log.e(TAG, "Timestamps were not verified successfully");
            return;
        }

        byte[] DET = new byte[Constants.DET_SIZE];
        linkData.get(DET);

        if(!ac.isDET(DET)) {
            Log.e(TAG, "This Link was sent by a drone with the wrong DET");
            return;
        }

        byte[] HI = new byte[Constants.HI_SIZE];
        linkData.get(HI);

        byte[] parentDET = new byte[Constants.DET_SIZE];
        linkData.get(parentDET);

        if(!Arrays.equals(parentDET,HDA_DET)) {
            Log.e(TAG, "The parent DET does not match with the cached HDA DET");
            return;
        }

        byte[] signature = new byte[Constants.SIGNATURE_SIZE];
        linkData.get(signature);

        boolean signatureValid = Ed25519.verify(
                signature, 0, HDA_PUB_KEY, 0, entire_message, 0, messageSize);

        if(!signatureValid) {
            Log.e(TAG, "The signature verification was not successful");
            StringWriter sw = new StringWriter();
            for(byte b : entire_message) {
                sw.append(String.format("%02X ", b));
            }
            Log.d(TAG, "Entire message: " + sw);
            return;
        }

        Link link = new Link();
        link.setPublicKey(HI);
        link.setMsgCounter(auth.getMsgCounter());
        link.setTimestamp(auth.getTimestamp());
        ac.link.postValue(link);

        ac.setPublicKey(HI);
    }

    private static void handleEndorsementStructure(ByteBuffer endorsementStructureData, SAMType type, AircraftObject ac, AuthenticationData auth, OpenDroneIdDataManager dataManager) {
        int messageSize = endorsementStructureData.remaining() - SIGNATURE_SIZE;
        Log.d(TAG,"Started handling endorsement structure (" + messageSize + ")");
        byte[] entire_message = new byte[messageSize];
        endorsementStructureData.slice().get(entire_message);

        int VNB = endorsementStructureData.getInt();
        int VNA = endorsementStructureData.getInt();

        if(!verifyTimestamps(VNB, VNA)) {
            Log.e(TAG, "Timestamps were not verified successfully");
            return;
        }

        int evidenceSize = endorsementStructureData.remaining() - SIGNATURE_SIZE - DET_SIZE;
        ByteBuffer evidence = endorsementStructureData.slice();
        evidence.order(ByteOrder.LITTLE_ENDIAN).limit(evidenceSize);

        // Advance past the evidence
        endorsementStructureData.position(endorsementStructureData.position() + evidenceSize);

        byte[] DET = new byte[DET_SIZE];
        endorsementStructureData.get(DET);

        byte[] publicKey = ac.getPublicKey();

        if(!ac.isDET(DET) || publicKey == null) {
            Log.e(TAG, "The DET is incorrect or the public key is still null");
            return;
        }

        byte[] signature = new byte[Constants.SIGNATURE_SIZE];
        endorsementStructureData.get(signature);

        boolean signatureValid = Ed25519.verify(
                signature, 0, publicKey, 0, entire_message, 0, messageSize);

        if(!signatureValid) {
            Log.e(TAG, "The signature verification was not successful");
            StringWriter sw = new StringWriter();
            for(byte b : entire_message) {
                sw.append(String.format("%02X ", b));
            }
            Log.d(TAG, "Entire message: " + sw);
            return;
        }

        switch (type) {
            case Wrapper:
                handleWrapper(evidence, auth, dataManager, ac);
                break;
            case Manifest:
                handleManifest(evidence, auth, ac);
                break;
            default:
        }
    }

    private static void handleWrapper(ByteBuffer wrapperData, AuthenticationData auth, OpenDroneIdDataManager dataManager, AircraftObject ac) {
        if ((wrapperData.remaining() % 25) != 0) {
            Log.e(TAG, "Wrapper sanity check failed");
            return;
        }
        int messageCount = wrapperData.remaining() / 25;

        Log.d(TAG,"Started handling wrapper (" + messageCount + ")");

        Connection connection = ac.connection.getValue();
        if(connection == null) {
            Log.e(TAG, "Connection is null");
            return;
        }

        OpenDroneIdParser.Type[] messageTypes = new OpenDroneIdParser.Type[messageCount];

        for(int i = 0; i < messageCount; i++) {
            byte[] message = new byte[25];
            wrapperData.get(message);
            messageTypes[i] = dataManager.receiveVerifiedData(message, auth.getTimestamp(), connection.macAddress, ac.getMacAddress(), connection.rssi, connection.transportType);
        }

        Wrapper wrapper = new Wrapper(messageCount);
        wrapper.setMessageTypes(messageTypes);
        wrapper.setTimestamp(auth.getTimestamp());
        wrapper.setMsgCounter(auth.getMsgCounter());
        ac.wrapper.postValue(wrapper);
    }

    private static void handleManifest(ByteBuffer manifestData, AuthenticationData auth, AircraftObject ac) {
        int hashLength = 8;
        if ((manifestData.remaining() % hashLength) != 0) {
            return;
        }
        int hashCount = (manifestData.remaining() / hashLength) - 2;

        // TODO: Check previous hash and store current hash

        byte[] previousHash = new byte[Constants.HASH_SIZE];
        manifestData.get(previousHash);

        byte[] currentHash = new byte[Constants.HASH_SIZE];
        manifestData.get(currentHash);

        ArrayList<OpenDroneIdParser.Type> messageTypes = new ArrayList<>();

        for(int i = 0; i < hashCount; i++) {
            byte[] messageHash = new byte[HASH_SIZE];
            manifestData.get(messageHash);

            // There is probably a better way of doing this by changing the way the messages are stored

            Identification identification1 = ac.identification1.getValue();
            if(identification1 != null && !identification1.isVerified()) {
                if (Arrays.equals(identification1.getHash(), messageHash)) {
                    identification1.setVerified(true);
                    ac.identification1.postValue(identification1);
                    messageTypes.add(OpenDroneIdParser.Type.BASIC_ID);
                }
            }

            Identification identification2 = ac.identification2.getValue();
            if(identification2 != null && !identification2.isVerified()) {
                if (Arrays.equals(identification2.getHash(), messageHash)) {
                    identification2.setVerified(true);
                    ac.identification2.postValue(identification2);
                    messageTypes.add(OpenDroneIdParser.Type.BASIC_ID);
                }
            }

            LocationData location = ac.location.getValue();
            if(location != null && !location.isVerified()) {
                if (Arrays.equals(location.getHash(), messageHash)) {
                    location.setVerified(true);
                    ac.location.postValue(location);
                    messageTypes.add(OpenDroneIdParser.Type.LOCATION);
                }
            }

            SelfIdData selfid = ac.selfid.getValue();
            if(selfid != null && !selfid.isVerified()) {
                if (Arrays.equals(selfid.getHash(), messageHash)) {
                    selfid.setVerified(true);
                    ac.selfid.postValue(selfid);
                    messageTypes.add(OpenDroneIdParser.Type.SELFID);
                }
            }

            SystemData system = ac.system.getValue();
            if(system != null && !system.isVerified()) {
                if (Arrays.equals(system.getHash(), messageHash)) {
                    system.setVerified(true);
                    ac.system.postValue(system);
                    messageTypes.add(OpenDroneIdParser.Type.SYSTEM);
                }
            }

            OperatorIdData operatorid = ac.operatorid.getValue();
            if(operatorid != null && !operatorid.isVerified()) {
                if (Arrays.equals(operatorid.getHash(), messageHash)) {
                    operatorid.setVerified(true);
                    ac.operatorid.postValue(operatorid);
                    messageTypes.add(OpenDroneIdParser.Type.OPERATOR_ID);
                }
            }
        }

        Manifest manifest = new Manifest();
        manifest.setMessageTypes(messageTypes.toArray(new OpenDroneIdParser.Type[0]));
        manifest.setTimestamp(auth.getTimestamp());
        manifest.setMsgCounter(auth.getMsgCounter());
        ac.manifest.postValue(manifest);
    }

    private static boolean verifyTimestamps(int ValidNotBeforeTimestamp, int ValidNotAfterTimestamp) {
        Timestamp VNB = new Timestamp((Constants.TIMESTAMP_OFFSET + ValidNotBeforeTimestamp) * 1000L);
        Timestamp VNA = new Timestamp((Constants.TIMESTAMP_OFFSET + ValidNotAfterTimestamp) * 1000L);

        Timestamp now = new Timestamp(System.currentTimeMillis());

        if(VNB.after(now)) {
            Log.e(TAG, "VNB timestamp is after current time");
            return false;
        }
        if(VNA.before(now)) {
            Log.e(TAG, "VNA timestamp (" + VNA + ") is already expired");
            return false;
        }
        return true;
    }

}