/*
 * Copyright (C) 2019 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 */
package org.securedroneid.android.data;

import android.util.Log;

import androidx.annotation.NonNull;
import androidx.lifecycle.MutableLiveData;

import org.securedroneid.android.Constants;
import org.securedroneid.android.bluetooth.DRIPDataManager;
import org.securedroneid.android.bluetooth.OpenDroneIdDataManager;
import org.xbill.DNS.Lookup;
import org.xbill.DNS.Record;
import org.xbill.DNS.SimpleResolver;
import org.xbill.DNS.Type;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;

public class AircraftObject {
    final public MutableLiveData<Connection> connection = new MutableLiveData<>();
    final public MutableLiveData<Identification> identification1 = new MutableLiveData<>();
    final public MutableLiveData<Identification> identification2 = new MutableLiveData<>();
    final public MutableLiveData<Identification> id1Shadow = new MutableLiveData<>();
    final public MutableLiveData<Identification> id2Shadow = new MutableLiveData<>();
    final public MutableLiveData<LocationData> location = new MutableLiveData<>();
    final public MutableLiveData<Manifest> manifest = new MutableLiveData<>();
    final public MutableLiveData<Wrapper> wrapper = new MutableLiveData<>();
    final public MutableLiveData<Link> link = new MutableLiveData<>();
    final public MutableLiveData<AuthenticationData> authentication = new MutableLiveData<>();
    final public MutableLiveData<SelfIdData> selfid = new MutableLiveData<>();
    final public MutableLiveData<SystemData> system = new MutableLiveData<>();
    final public MutableLiveData<OperatorIdData> operatorid = new MutableLiveData<>();
    final public MutableLiveData<DNSData> dns = new MutableLiveData<>();

    private final long macAddress;
    private byte[] publicKey;

    public AircraftObject(long macAddress) {
        this.macAddress = macAddress;
    }
    public long getMacAddress() { return macAddress; }

    public Connection getConnection() { return connection.getValue(); }
    public Identification getIdentification1() { return identification1.getValue(); }
    public Identification getIdentification2() { return identification2.getValue(); }
    public LocationData getLocation() { return location.getValue(); }
    public Manifest getManifest() { return manifest.getValue(); }
    public Wrapper getWrapper() { return wrapper.getValue(); }
    public Link getLink() { return link.getValue(); }
    public AuthenticationData getAuthentication() { return authentication.getValue(); }
    public SelfIdData getSelfID() { return selfid.getValue(); }
    public SystemData getSystem() { return system.getValue(); }
    public OperatorIdData getOperatorID() { return operatorid.getValue(); }

    // Non-zero authentication data pages do not contain the following fields. Save them for displaying
    private int authLastPageIndexSave;
    private int authLengthSave;
    private long authTimestampSave;

    // Multiple authentication messages are possible, each transmitting a part of the
    // authentication signature. Collect the data into authDataCombined.
    private final byte[] authDataCombined = new byte[Constants.MAX_AUTH_DATA];

    public AuthenticationData combineAuthentication(AuthenticationData newData, OpenDroneIdDataManager dataManager) { // TODO: check that the msg version is the same before combining
        AuthenticationData currData = authentication.getValue();
        if (currData == null)
            currData = new AuthenticationData();

        currData.setMsgCounter(newData.getMsgCounter());
        currData.setTimestamp(newData.getTimestamp());
        currData.setMsgVersion(newData.getMsgVersion());

        int offset = 0;
        int amount = Constants.MAX_AUTH_PAGE_ZERO_SIZE;
        if (newData.getAuthDataPage() == 0)  {
            authLastPageIndexSave = newData.getAuthLastPageIndex();
            authLengthSave = newData.getAuthLength();
            authTimestampSave = newData.getAuthTimestamp();
        } else {
            offset = Constants.MAX_AUTH_PAGE_ZERO_SIZE + (newData.getAuthDataPage() - 1) * Constants.MAX_AUTH_PAGE_NON_ZERO_SIZE;
            amount = Constants.MAX_AUTH_PAGE_NON_ZERO_SIZE;
        }
        for (int i = offset; i < offset + amount; i++)
            authDataCombined[i] = newData.getAuthData()[i];

        currData.setAuthType(newData.getAuthType());
        currData.setAuthLastPageIndex(authLastPageIndexSave);
        currData.setAuthLength(authLengthSave);
        currData.setAuthTimestamp(authTimestampSave);
        currData.setAuthData(authDataCombined);

        if (authLastPageIndexSave == newData.getAuthDataPage() && currData.getAuthType() == AuthenticationData.AuthTypeEnum.Specific_Authentication){
            ByteBuffer byteBuffer = ByteBuffer.wrap(currData.getAuthData(), 0, currData.getAuthLength());
            byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
            authLastPageIndexSave = 0;
            authLengthSave = 0;
            authTimestampSave = 0;
            DRIPDataManager.handleSAMMessage(byteBuffer, this, currData, dataManager);
        }

        return currData;
    }

    private int idToShow = 0;

    // When two different BasicId messages have been received, use this function to force a periodic
    // swap between their uasId in the list view. It is assumed this is called once per second.
    // The change logic is slowed down to once per three seconds.
    public void updateShadowBasicId() {
        switch (idToShow) {
            case 0:
                id1Shadow.setValue(identification1.getValue());
                idToShow++;
                break;
            case 3:
                Identification id2 = identification2.getValue();
                if (id2 != null && id2.getIdType() != Identification.IdTypeEnum.None)
                    id2Shadow.setValue(identification2.getValue());
                idToShow++;
                break;
            case 6:
                idToShow = 0;
                break;
            default:
                idToShow++;
        }
    }

    public void dnsLookup(Identification id) {
        final String domain = buildFqdn(id.getUasId());
        try {
            Log.d("AircraftObject", "Started lookup on: " + domain);
            Lookup lookup = new Lookup(domain, Type.TXT);
            SimpleResolver resolver = new SimpleResolver("8.8.8.8");
            lookup.setResolver(resolver);
            new Thread(){
                public void run(){
                    lookup.run();
                    if (lookup.getResult() == Lookup.SUCCESSFUL) {
                        Record[] records = lookup.getAnswers();
                        Log.d("AircraftObject", "Received: " + records[0].rdataToString());
                        String data = records[0].rdataToString().replaceAll("[\\\"\\\\]", "");
                        String owner = data.split(",")[0];
                        owner = owner.substring(owner.indexOf("=") + 1);
                        String model = data.split(",")[1];
                        model = model.substring(model.indexOf("=") + 1);
                        DNSData dnsData = new DNSData(owner, model);
                        dns.postValue(dnsData);
                    } else {
                        Log.e("AircraftObject", "ERROR: No TXT record found for " + domain + ". Error: " + lookup.getErrorString());
                    }
                }
            }.start();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private String buildFqdn(byte[] DET) {

        String apex = "vikingssl.com"; // Change this based on the implementation of the DNS
        int position = 0;

        String prefix = readBits(DET, position, 28);
        position += 28;

        String raa = readBits(DET, position, 14);
        position += 14;

        String hda = readBits(DET, position, 14);
        position += 14;

        String oga = readBits(DET, position, 8);
        position += 8;

        String id = readBits(DET, position, 64);
        position += 64;

        String result = id + "." + oga + "." + hda + "." + raa + "." + prefix + "." + apex;

        return result;
    }

    public static String readBits(byte[] data, int offset, int numBits) {
        StringBuilder result = new StringBuilder();
        int bitIndex = 0;

        for (int i = offset; bitIndex < numBits; i++) {
            int byteIndex = i / 8;
            int bitOffset = i % 8;

            int mask = 1 << (7 - bitOffset); // Create a mask for the current bit

            if ((data[byteIndex] & mask) != 0) {
                result.append("1");
            } else {
                result.append("0");
            }

            bitIndex++;
        }

        return binaryToHex(result.toString());
    }

    // Convert binary string to hexadecimal
    public static String binaryToHex(String binaryString) {
        int remainder = binaryString.length() % 4;
        String output = "";
        if(remainder != 0) {
            int desiredLength = binaryString.length() + remainder;
            binaryString = String.format("%" + desiredLength + "s", binaryString).replace(' ', '0');
        }
        for (int i = 0; i < binaryString.length(); i += 4) {
            String substring = binaryString.substring(i, i + 4);
            output += Integer.toString(Integer.parseInt(substring, 2), 16);
        }
        return output;
    }

    @Override @NonNull
    public String toString() {
        return "AircraftObject{" +
                "macAddress=" + macAddress +
                ", identification1=" + identification1 +
                ", identification2=" + identification2 +
                '}';
    }

    public boolean isDET(byte[] DET) {
        Identification id1 = identification1.getValue();
        Identification id2 = identification2.getValue();

        if(id1 == null || id2 == null) {
            return false;
        }

        if(id1.getIdType() == Identification.IdTypeEnum.Specific_Session_ID) {
            return Arrays.equals(DET,id1.getUasId());
        } else if(id2.getIdType() == Identification.IdTypeEnum.Specific_Session_ID) {
            return Arrays.equals(DET,id2.getUasId());
        }

        return false;
    }

    public byte[] getPublicKey() {
        return this.publicKey;
    }

    public void setPublicKey(byte[] pubKey) {
        this.publicKey = pubKey;
    }


}
